@@ -4,58 +4,39 @@ Python package
 
 from __future__ import annotations
 
+import atexit
 import base64
 import binascii
 import itertools
 import json
 import mimetypes
 import multiprocessing
+import os
 import re
+import threading
+import time
 import uuid
+from collections.abc import (
+    Callable,
+    Mapping,
+)
+from contextlib import AbstractContextManager
+from dataclasses import (
+    asdict,
+    is_dataclass,
+)
+from datetime import datetime
 from multiprocessing.process import BaseProcess
 from pathlib import Path
 from types import TracebackType
-from urllib.parse import quote_plus
-
-from neptune_scale.logging.console_log_capture import ConsoleLogCaptureThread
-from neptune_scale.sync.files import (
-    generate_destination,
-    guess_mime_type_from_bytes,
-    guess_mime_type_from_file,
-)
-from neptune_scale.sync.metadata_splitter import (
-    FileRefData,
-    MetadataSplitter,
-    Metrics,
-    datetime_to_proto,
-    histograms_to_update_run_snapshots,
-    make_step,
-    string_series_to_update_run_snapshots,
-)
-from neptune_scale.sync.operations_repository import (
-    FileUploadRequest,
-    OperationsRepository,
-)
-from neptune_scale.types import (
-    File,
-    Histogram,
-)
-
-__all__ = ["Run"]
-
-import atexit
-import os
-import threading
-import time
-from collections.abc import Callable
-from contextlib import AbstractContextManager
-from datetime import datetime
 from typing import (
     Any,
     Literal,
     Optional,
     Union,
+    cast,
 )
+from urllib.parse import quote_plus
 
 from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import ForkPoint
 from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import Run as CreateRun
@@ -73,22 +54,38 @@ from neptune_scale.exceptions import (
     NeptuneProjectNotProvided,
     NeptuneSynchronizationStopped,
 )
-from neptune_scale.sync.errors_tracking import (
-    ErrorsMonitor,
-    ErrorsQueue,
+from neptune_scale.logging.console_log_capture import ConsoleLogCaptureThread
+from neptune_scale.sync.errors_tracking import ErrorsMonitor
+from neptune_scale.sync.files import (
+    generate_destination,
+    guess_mime_type_from_bytes,
+    guess_mime_type_from_file,
 )
 from neptune_scale.sync.lag_tracking import LagTracker
+from neptune_scale.sync.metadata_splitter import (
+    FileRefData,
+    MetadataSplitter,
+    Metrics,
+    datetime_to_proto,
+    histograms_to_update_run_snapshots,
+    make_step,
+    string_series_to_update_run_snapshots,
+)
+from neptune_scale.sync.operations_repository import (
+    FileUploadRequest,
+    OperationsRepository,
+)
 from neptune_scale.sync.parameters import (
     MAX_EXPERIMENT_NAME_LENGTH,
     MAX_RUN_ID_LENGTH,
-    MINIMAL_WAIT_FOR_ACK_SLEEP_TIME,
-    MINIMAL_WAIT_FOR_PUT_SLEEP_TIME,
     OPERATION_REPOSITORY_POLL_SLEEP_TIME,
-    STOP_MESSAGE_FREQUENCY,
 )
-from neptune_scale.sync.sequence_tracker import SequenceTracker
 from neptune_scale.sync.supervisor import ProcessSupervisor
 from neptune_scale.sync.sync_process import run_sync_process
+from neptune_scale.types import (
+    File,
+    Histogram,
+)
 from neptune_scale.util import envs
 from neptune_scale.util.envs import (
     API_TOKEN_ENV_NAME,
@@ -96,13 +93,14 @@ from neptune_scale.util.envs import (
     PROJECT_ENV_NAME,
 )
 from neptune_scale.util.generate_run_id import generate_run_id
-from neptune_scale.util.logger import get_logger
-from neptune_scale.util.shared_var import (
-    SharedFloat,
-    SharedInt,
+from neptune_scale.util.logger import (
+    ThrottledLogger,
+    get_logger,
 )
 from neptune_scale.util.timer import Timer
 
+__all__ = ["Run"]
+
 logger = get_logger()
 
 
@@ -159,6 +157,7 @@ class Run(AbstractContextManager):
             on_async_lag_callback: Callback function triggered when the duration between the queueing and synchronization
             on_queue_full_callback: Callback function triggered when the queue is full. The function should take the exception
                 that made the queue full as its argument and an optional timestamp of the last time the exception was raised.
+                Deprecated.
             on_network_error_callback: Callback function triggered when a network error occurs.
             on_error_callback: The default callback function triggered when error occurs. It applies if an error
                 wasn't caught by other callbacks.
@@ -265,7 +264,6 @@ class Run(AbstractContextManager):
                 raise NeptuneDatabaseConflict(path=operations_repository_path.name)
             self._operations_repo.save_metadata(self._project, self._run_id)
 
-            self._sequence_tracker: Optional[SequenceTracker] = SequenceTracker()
             self._console_log_capture: Optional[ConsoleLogCaptureThread] = (
                 None
                 if not enable_console_log_capture
@@ -283,30 +281,23 @@ class Run(AbstractContextManager):
         else:
             self._storage_directory_path = None
             self._operations_repo = None
-            self._sequence_tracker = None
             self._console_log_capture = None
             self._logging_enabled = False
 
         if mode == "async":
-            assert self._sequence_tracker is not None
             if self._api_token is None:
                 raise NeptuneApiTokenNotProvided()
+            assert self._operations_repo is not None
 
             spawn_mp_context = multiprocessing.get_context("spawn")
 
-            self._errors_queue: Optional[ErrorsQueue] = ErrorsQueue(spawn_mp_context)
             self._errors_monitor: Optional[ErrorsMonitor] = ErrorsMonitor(
-                errors_queue=self._errors_queue,
-                on_queue_full_callback=on_queue_full_callback,
+                operations_repository=self._operations_repo,
                 on_network_error_callback=on_network_error_callback,
                 on_error_callback=on_error_callback,
                 on_warning_callback=on_warning_callback,
             )
 
-            self._last_queued_seq: Optional[SharedInt] = SharedInt(spawn_mp_context, -1)
-            self._last_ack_seq: Optional[SharedInt] = SharedInt(spawn_mp_context, -1)
-            self._last_ack_timestamp: Optional[SharedFloat] = SharedFloat(spawn_mp_context, -1)
-
             self._sync_process: Optional[BaseProcess] = spawn_mp_context.Process(
                 name="SyncProcess",
                 target=run_sync_process,
@@ -314,11 +305,7 @@ class Run(AbstractContextManager):
                     "project": self._project,
                     "family": self._run_id,
                     "operations_repository_path": operations_repository_path,
-                    "errors_queue": self._errors_queue,
                     "api_token": self._api_token,
-                    "last_queued_seq": self._last_queued_seq,
-                    "last_ack_seq": self._last_ack_seq,
-                    "last_ack_timestamp": self._last_ack_timestamp,
                 },
             )
             self._sync_process_supervisor: Optional[ProcessSupervisor] = ProcessSupervisor(
@@ -328,8 +315,7 @@ class Run(AbstractContextManager):
             self._lag_tracker: Optional[LagTracker] = None
             if async_lag_threshold is not None and on_async_lag_callback is not None:
                 self._lag_tracker = LagTracker(
-                    sequence_tracker=self._sequence_tracker,
-                    last_ack_timestamp=self._last_ack_timestamp,
+                    operations_repository=self._operations_repo,
                     async_lag_threshold=async_lag_threshold,
                     on_async_lag_callback=on_async_lag_callback,
                 )
@@ -353,11 +339,7 @@ class Run(AbstractContextManager):
                     raise RuntimeError(message) from e
                 self._sync_process_supervisor.start()
         else:
-            self._errors_queue = None
             self._errors_monitor = None
-            self._last_queued_seq = None
-            self._last_ack_seq = None
-            self._last_ack_timestamp = None
             self._sync_process = None
             self._sync_process_supervisor = None
             self._lag_tracker = None
@@ -372,6 +354,9 @@ class Run(AbstractContextManager):
                 fork_step=fork_step,
             )
 
+            if mode == "async":
+                self._print_run_info()
+
         # enable stdout / stderr logging at the end of the constructor, when everything is ready
         if self._console_log_capture is not None:
             self._console_log_capture.start()
@@ -379,8 +364,14 @@ class Run(AbstractContextManager):
     def _handle_sync_process_death(self) -> None:
         with self._lock:
             if not self._is_closing:
-                if self._errors_queue is not None:
-                    self._errors_queue.put(NeptuneSynchronizationStopped())
+                if self._operations_repo is not None:
+                    self._operations_repo.save_errors(errors=[NeptuneSynchronizationStopped()])
+
+    def _print_run_info(self) -> None:
+        try:
+            logger.info(f"View your run in the app:\n{self.get_run_url()}")
+        except Exception as e:
+            logger.debug(f"Error while printing run info: {e}")
 
     def _close(self, *, timeout: Optional[float] = None) -> None:
         timer = Timer(timeout)
@@ -399,16 +390,16 @@ class Run(AbstractContextManager):
 
             logger.debug(f"Run is closing, timeout={timeout}")
 
+        if self._sync_process_supervisor is not None:
+            self._sync_process_supervisor.interrupt()
+            self._sync_process_supervisor.join(timeout=timer.remaining_time())
+
         if self._sync_process is not None and self._sync_process.is_alive():
             self.wait_for_processing(timeout=timer.remaining_time())
 
             self._sync_process.terminate()
             self._sync_process.join(timeout=timer.remaining_time())
 
-        if self._sync_process_supervisor is not None:
-            self._sync_process_supervisor.interrupt()
-            self._sync_process_supervisor.join(timeout=timer.remaining_time())
-
         if self._lag_tracker is not None:
             self._lag_tracker.interrupt()
             self._lag_tracker.join(timeout=timer.remaining_time())
@@ -424,9 +415,6 @@ class Run(AbstractContextManager):
         if self._operations_repo is not None:
             self._operations_repo.close(cleanup_files=True)
 
-        if self._errors_queue is not None:
-            self._errors_queue.close()
-
         if self._storage_directory_path is not None:
             try:
                 logger.debug(f"Removing directory {self._storage_directory_path}")
@@ -497,7 +485,7 @@ class Run(AbstractContextManager):
         fork_run_id: Optional[str],
         fork_step: Optional[Union[int, float]],
     ) -> None:
-        if self._operations_repo is None or self._sequence_tracker is None:
+        if self._operations_repo is None:
             logger.debug("Run is in mode that doesn't support creating runs.")
             return
 
@@ -516,8 +504,7 @@ class Run(AbstractContextManager):
             creation_time=None if creation_time is None else datetime_to_proto(creation_time),
         )
 
-        sequence_id = self._operations_repo.save_create_run(create_run)
-        self._sequence_tracker.update_sequence_id(sequence_id)
+        self._operations_repo.save_create_run(create_run)
 
     def log_metrics(
         self,
@@ -579,27 +566,76 @@ class Run(AbstractContextManager):
             ),
         )
 
+    @staticmethod
+    def _flatten(data: Any) -> dict[str, Any]:
+        flattened = {}
+
+        def _flatten_inner(d: Any, prefix: str = "") -> None:
+            if is_dataclass(d):
+                d = asdict(d)  # type: ignore
+            if not isinstance(d, Mapping):
+                raise TypeError(f"Cannot flatten value of type {type(d)}. Try `flatten=False`.")
+            for key, value in d.items():
+                str_key = str(key)
+                new_key = f"{prefix}/{str_key}" if prefix else str_key
+                if isinstance(value, Mapping) or is_dataclass(value):
+                    _flatten_inner(value, prefix=new_key)
+                else:
+                    flattened[new_key] = value
+
+        _flatten_inner(data)
+        return flattened
+
+    @staticmethod
+    def _cast_unsupported(
+        data: Any,
+    ) -> dict[str, Union[str, float, int, bool, datetime, list[str], set[str], tuple[str, ...]]]:
+        result: dict[str, Union[str, float, int, bool, datetime, list[str], set[str], tuple[str, ...]]] = {}
+
+        for k, v in data.items():
+            if (
+                isinstance(v, (float, bool, int, str, datetime))
+                or isinstance(v, (list, set, tuple))
+                and all(isinstance(item, str) for item in v)
+            ):
+                result[k] = v
+            else:
+                result[k] = "" if v is None else str(v)  # If value is None, store as empty string
+        return result
+
     def log_configs(
         self,
-        data: Optional[dict[str, Union[float, bool, int, str, datetime, list, set, tuple]]] = None,
+        data: Optional[
+            Union[
+                Mapping[
+                    str,
+                    Union[str, float, int, bool, datetime, list[Any], set[Any], tuple[Any, ...]],
+                ],
+                Any,
+            ]
+        ],
+        flatten: bool = False,
+        cast_unsupported: bool = False,
     ) -> None:
         """
         Logs the specified metadata to a Neptune run.
 
-        You can log configurations or other single values. Pass the metadata as a dictionary {key: value} with
-
-        - key: path to where the metadata should be stored in the run.
-        - value: configuration or other single value to log.
+        You can log configurations or other single values.
+        Pass the data as a dataclass or a mapping-like object, where:
+        - keys are strings specifying the path to the metadata in the run structure.
+        - values are the configurations or other single values to log.
 
         For example, {"parameters/learning_rate": 0.001}.
         In the attribute path, each forward slash "/" nests the attribute under a namespace.
         Use namespaces to structure the metadata into meaningful categories.
 
         Args:
-            data: Dictionary of configs or other values to log.
+            data: Configs or other values to log, as a `dict`, dataclass, or other mapping-like object with an `.items()` method.
                 Available types: float, integer, Boolean, string, and datetime.
+                Any `datetime` values that don't have the `tzinfo` attribute set are assumed to be in the local timezone.
+            flatten: Flattens nested dictionaries and dataclasses before logging. Defaults to False.
+            cast_unsupported: Casts unsupported types to strings before logging. Defaults to False.
 
-        Any `datetime` values that don't have the `tzinfo` attribute set are assumed to be in the local timezone.
 
         Example:
             ```
@@ -614,6 +650,26 @@ class Run(AbstractContextManager):
                 )
             ```
         """
+        if data is None:
+            return
+
+        if is_dataclass(data):
+            data = asdict(data)  # type: ignore
+        elif not isinstance(data, Mapping):
+            raise TypeError(
+                f"configs must be a mapping-like object with an `.items()` method (e.g., `dict`) or a `dataclass` instance, or `NoneType` (was {type(data)})"
+            )
+
+        if flatten:
+            data = Run._flatten(data)
+
+        if cast_unsupported:
+            data = Run._cast_unsupported(data)
+
+        data = cast(
+            dict[str, Union[str, float, int, bool, datetime, list[str], set[str], tuple[str, ...]]],
+            data,
+        )
         self._log(configs=data)
 
     def log_string_series(
@@ -659,7 +715,11 @@ class Run(AbstractContextManager):
         self._log(timestamp=timestamp, step=step, string_series=data)
 
     def log_histograms(
-        self, histograms: dict[str, Histogram], step: Union[float, int], *, timestamp: Optional[datetime] = None
+        self,
+        histograms: dict[str, Histogram],
+        step: Union[float, int],
+        *,
+        timestamp: Optional[datetime] = None,
     ) -> None:
         """Logs the specified histograms at a particular step.
 
@@ -804,7 +864,9 @@ class Run(AbstractContextManager):
         self,
         step: Optional[Union[float, int]] = None,
         timestamp: Optional[datetime] = None,
-        configs: Optional[dict[str, Union[float, bool, int, str, datetime, list, set, tuple]]] = None,
+        configs: Optional[
+            dict[str, Union[str, float, int, bool, datetime, list[str], set[str], tuple[str, ...]]]
+        ] = None,
         metrics: Optional[dict[str, Union[float, int]]] = None,
         tags_add: Optional[dict[str, Union[list[str], set[str], tuple[str]]]] = None,
         tags_remove: Optional[dict[str, Union[list[str], set[str], tuple[str]]]] = None,
@@ -831,7 +893,9 @@ class Run(AbstractContextManager):
         self,
         timestamp: Optional[datetime] = None,
         step: Optional[Union[float, int]] = None,
-        configs: Optional[dict[str, Union[float, bool, int, str, datetime, list, set, tuple]]] = None,
+        configs: Optional[
+            dict[str, Union[str, float, int, bool, datetime, list[str], set[str], tuple[str, ...]]]
+        ] = None,
         metrics: Optional[Metrics] = None,
         files: Optional[dict[str, Union[str, Path, bytes, File]]] = None,
         string_series: Optional[dict[str, str]] = None,
@@ -885,7 +949,6 @@ class Run(AbstractContextManager):
             return
 
         assert self._operations_repo is not None
-        assert self._sequence_tracker is not None
 
         if timestamp is None:
             timestamp = datetime.now()
@@ -931,8 +994,7 @@ class Run(AbstractContextManager):
         if all_file_upload_requests:
             self._operations_repo.save_file_upload_requests([req for _, req in all_file_upload_requests])
 
-        sequence_id = self._operations_repo.save_update_run_snapshots(operations)
-        self._sequence_tracker.update_sequence_id(sequence_id)
+        self._operations_repo.save_update_run_snapshots(operations)
 
     @staticmethod
     def _file_request_to_file_ref_data(file_request: FileUploadRequest) -> FileRefData:
@@ -943,7 +1005,9 @@ class Run(AbstractContextManager):
         )
 
     def _prepare_files_for_upload(
-        self, files: Optional[dict[str, Union[str, Path, bytes, File]]], step: Optional[Union[float, int]] = None
+        self,
+        files: Optional[dict[str, Union[str, Path, bytes, File]]],
+        step: Optional[Union[float, int]] = None,
     ) -> list[tuple[str, FileUploadRequest]]:
         """Process user input to produce a list of (attribute-name, FileUploadRequest tuples)
 
@@ -959,7 +1023,12 @@ class Run(AbstractContextManager):
         for attr_name, file in files.items():
             try:
                 if isinstance(file, File):
-                    source, mime_type, destination, size = file.source, file.mime_type, file.destination, file.size
+                    source, mime_type, destination, size = (
+                        file.source,
+                        file.mime_type,
+                        file.destination,
+                        file.size,
+                    )
                 else:
                     source = file
                     mime_type = size = destination = None
@@ -1002,85 +1071,7 @@ class Run(AbstractContextManager):
 
         return result
 
-    def _wait(
-        self,
-        phrase: str,
-        sleep_time: float,
-        wait_seq: Optional[SharedInt],
-        timeout: Optional[float] = None,
-        verbose: bool = True,
-    ) -> None:
-        if timeout is not None and timeout <= 0:
-            return
-
-        if verbose:
-            logger.info(f"Waiting for all operations to be {phrase}")
-
-        if timeout is None and verbose:
-            logger.warning("No timeout specified. Waiting indefinitely")
-
-        timer = Timer(timeout)
-        wait_time = sleep_time
-        last_print_timestamp: Optional[float] = None
-
-        while True:
-            try:
-                with self._lock:
-                    if self._sync_process is None or not self._sync_process.is_alive():
-                        logger.warning("Waiting interrupted because sync process is not running")
-                        break
-
-                    assert wait_seq is not None
-                    assert self._sequence_tracker is not None
-
-                    # Handle the case where we get notified on `wait_seq` before we actually wait.
-                    # Otherwise, we would unnecessarily block, waiting on a notify_all() that never happens.
-                    if wait_seq.value >= self._sequence_tracker.last_sequence_id:
-                        break
-
-                wait_time = min(wait_time, timer.remaining_time_or_inf())
-                with wait_seq:
-                    wait_seq.wait(timeout=wait_time)
-                    value = wait_seq.value
-
-                last_queued_sequence_id = self._sequence_tracker.last_sequence_id
-
-                if value == -1:
-                    if last_queued_sequence_id != -1:
-                        last_print_timestamp = print_message(
-                            f"Waiting. No operations were {phrase} yet. Operations to sync: %s",
-                            last_queued_sequence_id,
-                            last_print=last_print_timestamp,
-                            verbose=verbose,
-                        )
-                    else:
-                        last_print_timestamp = print_message(
-                            f"Waiting. No operations were {phrase} yet",
-                            last_print=last_print_timestamp,
-                            verbose=verbose,
-                        )
-                elif value < last_queued_sequence_id:
-                    last_print_timestamp = print_message(
-                        f"Waiting for remaining %d operation(s) to be {phrase}",
-                        last_queued_sequence_id - value,
-                        last_print=last_print_timestamp,
-                        verbose=verbose,
-                    )
-                else:
-                    # Reaching the last queued sequence ID means that all operations were submitted
-                    if verbose:
-                        logger.info(f"All operations were {phrase}")
-                    break
-
-                if timer.is_expired():
-                    if verbose:
-                        logger.info("Waiting interrupted because timeout was reached")
-                    break
-            except KeyboardInterrupt:
-                if verbose:
-                    logger.warning("Waiting interrupted by user")
-
-    def wait_for_submission(self, timeout: Optional[float] = None, verbose: bool = True) -> None:
+    def wait_for_submission(self, timeout: Optional[float] = None, verbose: bool = True) -> bool:
         """
         Waits until all metadata is submitted to Neptune for processing.
 
@@ -1092,16 +1083,11 @@ class Run(AbstractContextManager):
             verbose (bool): If True (default), prints messages about the waiting process.
         """
         timer = Timer(timeout)
-        self._wait(
-            phrase="submitted",
-            sleep_time=MINIMAL_WAIT_FOR_PUT_SLEEP_TIME,
-            wait_seq=self._last_queued_seq,
-            timeout=timer.remaining_time(),
-            verbose=verbose,
-        )
-        self._wait_for_file_upload(timeout=timer.remaining_time(), verbose=verbose)
+        submission_status = self._wait_for_operation_submission(timeout=timer.remaining_time(), verbose=verbose)
+        files_status = self._wait_for_file_upload(timeout=timer.remaining_time(), verbose=verbose)
+        return submission_status and files_status
 
-    def wait_for_processing(self, timeout: Optional[float] = None, verbose: bool = True) -> None:
+    def wait_for_processing(self, timeout: Optional[float] = None, verbose: bool = True) -> bool:
         """
         Waits until all metadata is processed by Neptune.
 
@@ -1112,22 +1098,130 @@ class Run(AbstractContextManager):
             verbose (bool): If True (default), prints messages about the waiting process.
         """
         timer = Timer(timeout)
-        self._wait(
-            phrase="processed",
-            sleep_time=MINIMAL_WAIT_FOR_ACK_SLEEP_TIME,
-            wait_seq=self._last_ack_seq,
-            timeout=timer.remaining_time(),
-            verbose=verbose,
-        )
-        self._wait_for_file_upload(timeout=timer.remaining_time(), verbose=verbose)
+        processing_status = self._wait_for_operation_processing(timeout=timer.remaining_time(), verbose=verbose)
+        file_status = self._wait_for_file_upload(timeout=timer.remaining_time(), verbose=verbose)
+        return processing_status and file_status
+
+    def _wait_for_operation_submission(
+        self,
+        timeout: Optional[float] = None,
+        verbose: bool = True,
+    ) -> bool:
+        if self._operations_repo is None:
+            return True
+
+        if verbose:
+            logger.info("Waiting for all operations to be processed")
+
+        if timeout is None and verbose:
+            logger.warning("No timeout specified. Waiting indefinitely")
+
+        timer = Timer(timeout)
+        sleep_time = float(OPERATION_REPOSITORY_POLL_SLEEP_TIME)
+        throttled_logger = ThrottledLogger(logger, enabled=verbose)
+
+        while True:
+            try:
+                with self._lock:
+                    if self._sync_process is None or not self._sync_process.is_alive():
+                        logger.warning("Waiting interrupted because sync process is not running")
+                        timer = Timer(0)  # reset timer to avoid waiting indefinitely
+
+                # assumption: submissions are always behind or equal to the logged operations
+                submitted_sequence_id_range = self._operations_repo.get_operation_submission_sequence_id_range()
+                logged_sequence_id_range = self._operations_repo.get_operations_sequence_id_range()
+
+                if submitted_sequence_id_range is not None and logged_sequence_id_range is not None:
+                    _, max_submitted_sequence_id = submitted_sequence_id_range
+                    _, max_logged_sequence_id = logged_sequence_id_range
+                    operations_count = max(0, max_logged_sequence_id - max_submitted_sequence_id)
+
+                    if operations_count > 0:
+                        throttled_logger.info(
+                            "Waiting for remaining %d operation(s) to be submitted",
+                            operations_count,
+                        )
+                elif logged_sequence_id_range is not None:
+                    min_logged_sequence_id, max_logged_sequence_id = logged_sequence_id_range
+                    operations_count = max(0, max_logged_sequence_id - min_logged_sequence_id + 1)
+
+                    if operations_count > 0:
+                        throttled_logger.info(
+                            "Waiting. No operations were submitted yet. Operations to sync: %s",
+                            operations_count,
+                        )
+                else:
+                    operations_count = 0
+
+                if operations_count > 0:
+                    if timer.is_expired():
+                        logger.info("Waiting interrupted because timeout was reached")
+                        return False
+                    sleep_time = min(sleep_time, timer.remaining_time_or_inf())
+
+                    time.sleep(sleep_time)
+                else:
+                    logger.info("All operations were submitted")
+                    return True
+            except KeyboardInterrupt:
+                logger.warning("Waiting interrupted by user")
+                return False
+
+    def _wait_for_operation_processing(
+        self,
+        timeout: Optional[float] = None,
+        verbose: bool = True,
+    ) -> bool:
+        if self._operations_repo is None:
+            return True
+
+        if verbose:
+            logger.info("Waiting for all operations to be processed")
+
+        if timeout is None and verbose:
+            logger.warning("No timeout specified. Waiting indefinitely")
+
+        operations_count_limit = 10_000
+        timer = Timer(timeout)
+        sleep_time = float(OPERATION_REPOSITORY_POLL_SLEEP_TIME)
+        throttled_logger = ThrottledLogger(logger, enabled=verbose)
+
+        while True:
+            try:
+                with self._lock:
+                    if self._sync_process is None or not self._sync_process.is_alive():
+                        logger.warning("Waiting interrupted because sync process is not running")
+                        timer = Timer(0)  # reset timer to avoid waiting indefinitely
+
+                operations_count = self._operations_repo.get_operation_count(limit=operations_count_limit)
+
+                if operations_count > 0:
+                    throttled_logger.info(
+                        "Waiting for remaining %d%s operation(s) to be processed",
+                        operations_count,
+                        "" if operations_count < operations_count_limit else "+",
+                    )
+
+                    if timer.is_expired():
+                        logger.info("Waiting interrupted because timeout was reached")
+                        return False
+                    sleep_time = min(sleep_time, timer.remaining_time_or_inf())
+
+                    time.sleep(sleep_time)
+                else:
+                    logger.info("All operations were processed")
+                    return True
+            except KeyboardInterrupt:
+                logger.warning("Waiting interrupted by user")
+                return False
 
     def _wait_for_file_upload(
         self,
         timeout: Optional[float] = None,
         verbose: bool = True,
-    ) -> None:
-        if timeout is not None and timeout <= 0:
-            return
+    ) -> bool:
+        if self._operations_repo is None:
+            return True
 
         if verbose:
             logger.info("Waiting for all files to be uploaded")
@@ -1135,45 +1229,39 @@ class Run(AbstractContextManager):
         if timeout is None and verbose:
             logger.warning("No timeout specified. Waiting indefinitely")
 
-        upload_count_limit = 1_000_000
+        upload_count_limit = 10_000
         timer = Timer(timeout)
         sleep_time = float(OPERATION_REPOSITORY_POLL_SLEEP_TIME)
-        last_print_timestamp: Optional[float] = None
+        throttled_logger = ThrottledLogger(logger, enabled=verbose)
 
         while True:
             try:
                 with self._lock:
                     if self._sync_process is None or not self._sync_process.is_alive():
                         logger.warning("Waiting interrupted because sync process is not running")
-                        return
-                assert self._operations_repo is not None
+                        timer = Timer(0)  # reset timer to avoid waiting indefinitely
 
                 upload_count = self._operations_repo.get_file_upload_requests_count(limit=upload_count_limit)
 
                 if upload_count > 0:
-                    last_print_timestamp = print_message(
+                    throttled_logger.info(
                         "Waiting for remaining %d%s file(s) to be uploaded",
                         upload_count,
                         "" if upload_count < upload_count_limit else "+",
-                        last_print=last_print_timestamp,
-                        verbose=verbose,
                     )
 
                     if timer.is_expired():
-                        if verbose:
-                            logger.info("Waiting interrupted because timeout was reached")
-                        break
+                        logger.info("Waiting interrupted because timeout was reached")
+                        return False
                     sleep_time = min(sleep_time, timer.remaining_time_or_inf())
 
                     time.sleep(sleep_time)
                 else:
-                    if verbose:
-                        logger.info("All files were uploaded")
-                    break
+                    logger.info("All files were uploaded")
+                    return True
             except KeyboardInterrupt:
-                if verbose:
-                    logger.warning("Waiting interrupted by user")
-                return
+                logger.warning("Waiting interrupted by user")
+                return False
 
     def get_run_url(self) -> str:
         """
@@ -1194,16 +1282,6 @@ class Run(AbstractContextManager):
         return _get_run_url(self._api_token, self._project, experiment_name=self._experiment_name)
 
 
-def print_message(msg: str, *args: Any, last_print: Optional[float] = None, verbose: bool = True) -> Optional[float]:
-    current_time = time.time()
-
-    if verbose and (last_print is None or current_time - last_print > STOP_MESSAGE_FREQUENCY):
-        logger.info(msg, *args)
-        return current_time
-
-    return last_print
-
-
 def _sanitize_path_component(component: str) -> str:
     """Modify a path component to be safe for use in a filesystem:
     - Replace ASCII control characters, and characters invalid/troublesome on some filesystems with an underscore
@@ -1236,7 +1314,10 @@ def _resolve_run_storage_directory_path(
 
 
 def _get_run_url(
-    api_token: Optional[str], project_name: str, run_id: Optional[str] = None, experiment_name: Optional[str] = None
+    api_token: Optional[str],
+    project_name: str,
+    run_id: Optional[str] = None,
+    experiment_name: Optional[str] = None,
 ) -> str:
     assert not (run_id and experiment_name)
     if api_token is None:
