@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import pickle
 from pathlib import Path
 
 from neptune_scale.sync.parameters import (
@@ -7,10 +8,17 @@ from neptune_scale.sync.parameters import (
     OPERATION_REPOSITORY_TIMEOUT,
 )
 
-__all__ = ("OperationsRepository", "OperationType", "Operation", "Metadata", "SequenceId", "FileUploadRequest")
+__all__ = (
+    "OperationsRepository",
+    "OperationType",
+    "Operation",
+    "Metadata",
+    "SequenceId",
+    "FileUploadRequest",
+    "OperationSubmission",
+)
 
 import contextlib
-import datetime
 import os
 import sqlite3
 import threading
@@ -18,6 +26,7 @@ import time
 import typing
 from contextlib import AbstractContextManager
 from dataclasses import dataclass
+from datetime import datetime
 from enum import IntEnum
 from typing import (
     Literal,
@@ -30,7 +39,10 @@ from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import UpdateRunSnapshot
 
 from neptune_scale.exceptions import (
     NeptuneLocalStorageInUnsupportedVersion,
+    NeptuneScaleError,
+    NeptuneScaleWarning,
     NeptuneUnableToLogData,
+    NeptuneUnexpectedError,
 )
 from neptune_scale.util import (
     envs,
@@ -39,10 +51,12 @@ from neptune_scale.util import (
 
 logger = get_logger()
 
-DB_VERSION = "v3"
-BACKWARD_COMPATIBLE_DB_VERSIONS = ("v2", "v3")
+DB_VERSION = "v5"
+BACKWARD_COMPATIBLE_DB_VERSIONS = ("v5",)
 
+ErrorId = typing.NewType("ErrorId", int)
 SequenceId = typing.NewType("SequenceId", int)
+RequestId = typing.NewType("RequestId", str)
 
 
 class OperationType(IntEnum):
@@ -59,8 +73,66 @@ class Operation:
     operation_size_bytes: int
 
     @property
-    def ts(self) -> datetime.datetime:
-        return datetime.datetime.fromtimestamp(self.timestamp / 1000)
+    def ts(self) -> datetime:
+        return datetime.fromtimestamp(self.timestamp / 1000)
+
+
+@dataclass(frozen=True)
+class OperationSubmission:
+    sequence_id: SequenceId
+    timestamp: int
+    request_id: RequestId
+
+    @property
+    def ts(self) -> datetime:
+        return datetime.fromtimestamp(self.timestamp / 1000)
+
+
+@dataclass(frozen=True)
+class OperationError:
+    error_id: Optional[ErrorId]
+    timestamp: int
+    error_type: str
+    error_details: str
+    error_body: str
+    sequence_id: Optional[SequenceId] = None
+
+    @property
+    def ts(self) -> datetime:
+        return datetime.fromtimestamp(self.timestamp / 1000)
+
+    @staticmethod
+    def serialize_error(error: BaseException, sequence_id: Optional[SequenceId] = None) -> OperationError:
+        try:
+            if not isinstance(error, (NeptuneScaleError, NeptuneScaleWarning)):
+                serialized_error: BaseException = NeptuneUnexpectedError(reason=str(error))
+            else:
+                serialized_error = error
+            error_body = pickle.dumps(serialized_error).hex()
+        except (pickle.PicklingError, TypeError):
+            logger.error(f"Failed to serialize error {error}", exc_info=True)
+            error_body = ""
+
+        return OperationError(
+            error_id=None,  # This will be set when saving to the database
+            timestamp=int(time.time() * 1000),
+            error_type=type(error).__name__,
+            error_details=str(error),
+            error_body=error_body,
+            sequence_id=sequence_id,
+        )
+
+    def deserialize_error(self) -> BaseException:
+        try:
+            result = pickle.loads(bytes.fromhex(self.error_body))
+            if isinstance(result, (NeptuneScaleError, NeptuneScaleWarning)):
+                return result
+            else:
+                logger.error(f"Failed to deserialize error {self.error_type} {self.error_details}", exc_info=True)
+                return NeptuneUnexpectedError(reason=f"{self.error_type} {self.error_details}")
+        except (pickle.UnpicklingError, ValueError):
+            logger.error(f"Failed to deserialize error {self.error_type} {self.error_details}", exc_info=True)
+            return NeptuneUnexpectedError(reason=f"{self.error_type} {self.error_details}")
 
 
 @dataclass(frozen=True)
@@ -149,6 +221,16 @@ class OperationsRepository:
                 )"""
             )
 
+            conn.execute(
+                """
+                CREATE TABLE IF NOT EXISTS run_operation_submission (
+                    sequence_id INTEGER PRIMARY KEY,
+                    timestamp INTEGER NOT NULL,
+                    request_id TEXT NOT NULL
+                )
+                """
+            )
+
             conn.execute(
                 """
                 CREATE TABLE IF NOT EXISTS file_upload_requests (
@@ -162,6 +244,19 @@ class OperationsRepository:
                 """
             )
 
+            conn.execute(
+                """
+                CREATE TABLE IF NOT EXISTS operation_errors (
+                    error_id INTEGER PRIMARY KEY AUTOINCREMENT,
+                    timestamp INTEGER NOT NULL,
+                    error_type TEXT NOT NULL,
+                    error_details TEXT NOT NULL,
+                    error_body TEXT NOT NULL,
+                    sequence_id INTEGER
+                )
+                """
+            )
+
     def save_update_run_snapshots(self, updates: list[UpdateRunSnapshot]) -> SequenceId:
         """
         Guarantees:
@@ -248,7 +343,7 @@ class OperationsRepository:
         except NeptuneUnableToLogData:
             if self._log_failure_action == "raise":
                 raise
-            if self._log_failure_action == "drop":
+            else:
                 logger.error(f"Dropping {len(ops)} operations due to error", exc_info=True)
                 return None
 
@@ -319,6 +414,49 @@ class OperationsRepository:
                 # Return the number of rows affected
                 return cursor.rowcount or 0
 
+    def get_operation_count(self, limit: Optional[int] = None) -> int:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                return self._get_table_count(cursor, "run_operations", limit=limit)
+
+    def get_operations_sequence_id_range(self) -> Optional[tuple[SequenceId, SequenceId]]:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.execute(
+                    """
+                    SELECT MIN(sequence_id), MAX(sequence_id)
+                    FROM run_operations
+                    """
+                )
+
+                row = cursor.fetchone()
+                if not row:
+                    return None
+
+                min_seq_id, max_seq_id = row
+                if min_seq_id is None or max_seq_id is None:
+                    return None
+                return SequenceId(min_seq_id), SequenceId(max_seq_id)
+
+    def get_operations_min_timestamp(self) -> Optional[datetime]:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.execute(
+                    """
+                    SELECT timestamp
+                    FROM run_operations
+                    ORDER BY sequence_id ASC
+                    LIMIT 1
+                    """
+                )
+
+                row = cursor.fetchone()
+                if not row:
+                    return None
+
+                (timestamp,) = row
+                return datetime.fromtimestamp(timestamp / 1000)
+
     def save_metadata(self, project: str, run_id: str) -> None:
         with self._get_connection() as conn:  # type: ignore
             with contextlib.closing(conn.cursor()) as cursor:
@@ -363,25 +501,6 @@ class OperationsRepository:
 
                 return Metadata(project=project, run_id=run_id)
 
-    def get_sequence_id_range(self) -> Optional[tuple[SequenceId, SequenceId]]:
-        with self._get_connection() as conn:  # type: ignore
-            with contextlib.closing(conn.cursor()) as cursor:
-                cursor.execute(
-                    """
-                    SELECT MIN(sequence_id), MAX(sequence_id)
-                    FROM run_operations
-                    """
-                )
-
-                row = cursor.fetchone()
-                if not row:
-                    return None
-
-                min_seq_id, max_seq_id = row
-                if min_seq_id is None or max_seq_id is None:
-                    return None
-                return SequenceId(min_seq_id), SequenceId(max_seq_id)
-
     def save_file_upload_requests(self, files: list[FileUploadRequest]) -> SequenceId:
         with self._get_connection() as conn:  # type: ignore
             with contextlib.closing(conn.cursor()) as cursor:
@@ -440,6 +559,152 @@ class OperationsRepository:
             with contextlib.closing(conn.cursor()) as cursor:
                 return self._get_table_count(cursor, "file_upload_requests", limit=limit)
 
+    def save_operation_submissions(self, submissions: list[OperationSubmission]) -> SequenceId:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.executemany(
+                    """
+                    INSERT INTO run_operation_submission (sequence_id, timestamp, request_id)
+                    VALUES (?, ?, ?)
+                    """,
+                    [(status.sequence_id, status.timestamp, status.request_id) for status in submissions],
+                )
+                cursor.execute("SELECT last_insert_rowid()")
+                return SequenceId(cursor.fetchone()[0])
+
+    def get_operation_submissions(self, limit: int) -> list[OperationSubmission]:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.execute(
+                    """
+                    SELECT sequence_id, timestamp, request_id
+                    FROM run_operation_submission
+                    ORDER BY sequence_id ASC
+                    LIMIT ?
+                    """,
+                    (limit,),
+                )
+
+                rows = cursor.fetchall()
+        return [
+            OperationSubmission(
+                sequence_id=SequenceId(row[0]),
+                timestamp=row[1],
+                request_id=RequestId(row[2]),
+            )
+            for row in rows
+        ]
+
+    def delete_operation_submissions(self, up_to_seq_id: Optional[SequenceId]) -> int:
+        if up_to_seq_id is not None and up_to_seq_id <= 0:
+            return 0
+
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                if up_to_seq_id is None:
+                    cursor.execute("DELETE FROM run_operation_submission")
+                else:
+                    cursor.execute(
+                        "DELETE FROM run_operation_submission WHERE sequence_id <= ?",
+                        (up_to_seq_id,),
+                    )
+
+                return cursor.rowcount or 0
+
+    def get_operation_submission_count(self, limit: Optional[int] = None) -> int:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                return self._get_table_count(cursor, "run_operation_submission", limit=limit)
+
+    def get_operation_submission_sequence_id_range(self) -> Optional[tuple[SequenceId, SequenceId]]:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.execute(
+                    """
+                    SELECT MIN(sequence_id), MAX(sequence_id)
+                    FROM run_operation_submission
+                    """
+                )
+
+                row = cursor.fetchone()
+                if not row:
+                    return None
+
+                min_seq_id, max_seq_id = row
+                if min_seq_id is None or max_seq_id is None:
+                    return None
+                return SequenceId(min_seq_id), SequenceId(max_seq_id)
+
+    def get_errors(self, limit: int) -> list[OperationError]:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.execute(
+                    """
+                    SELECT error_id, timestamp, error_type, error_details,  error_body, sequence_id
+                    FROM operation_errors
+                    ORDER BY timestamp ASC
+                    LIMIT ?
+                    """,
+                    (limit,),
+                )
+
+                rows = cursor.fetchall()
+
+        return [
+            OperationError(
+                error_id=ErrorId(row[0]) if row[0] is not None else None,
+                timestamp=row[1],
+                error_type=row[2],
+                error_details=row[3],
+                error_body=row[4],
+                sequence_id=SequenceId(row[5]) if row[5] is not None else None,
+            )
+            for row in rows
+        ]
+
+    def save_errors(
+        self,
+        errors: typing.Iterable[BaseException],
+        sequence_id: Optional[SequenceId] = None,
+    ) -> ErrorId:
+        operation_errors = [OperationError.serialize_error(error, sequence_id) for error in errors]
+
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.executemany(
+                    """
+                    INSERT INTO operation_errors (timestamp, error_type, error_details, error_body, sequence_id)
+                    VALUES (?, ?, ?, ?, ?)
+                    """,
+                    [
+                        (error.timestamp, error.error_type, error.error_details, error.error_body, error.sequence_id)
+                        for error in operation_errors
+                    ],
+                )
+                cursor.execute("SELECT last_insert_rowid()")
+                return ErrorId(cursor.fetchone()[0])
+
+    def delete_errors(self, error_ids: typing.Iterable[ErrorId]) -> None:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.executemany(
+                    """
+                    DELETE FROM operation_errors
+                    WHERE error_id = ?
+                    """,
+                    [(error_id,) for error_id in error_ids],
+                )
+
+    def delete_all_errors(self) -> None:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                cursor.execute("DELETE FROM operation_errors")
+
+    def get_errors_count(self, limit: Optional[int] = None) -> int:
+        with self._get_connection() as conn:  # type: ignore
+            with contextlib.closing(conn.cursor()) as cursor:
+                return self._get_table_count(cursor, "operation_errors", limit=limit)
+
     def _is_repository_empty(self) -> bool:
         with self._get_connection() as conn:  # type: ignore
             with contextlib.closing(conn.cursor()) as cursor:
