@@ -1,18 +1,20 @@
 from __future__ import annotations
 
 import asyncio
-import logging
 from pathlib import Path
 
 import azure.core.exceptions
 from azure.core.pipeline.transport import AsyncioRequestsTransport
 from azure.storage.blob.aio import BlobClient
+from neptune_api.models import Provider
 
+from neptune_scale.sync.google_storage import upload_to_gcp
 from neptune_scale.sync.operations_repository import (
     FileUploadRequest,
     Metadata,
     Operation,
     OperationsRepository,
+    OperationSubmission,
     OperationType,
     SequenceId,
 )
@@ -23,13 +25,10 @@ __all__ = ("run_sync_process",)
 import datetime
 import functools as ft
 import os
-import queue
 import signal
 import threading
 from types import FrameType
 from typing import (
-    Generic,
-    NamedTuple,
     Optional,
     TypeVar,
 )
@@ -87,10 +86,8 @@ from neptune_scale.exceptions import (
 )
 from neptune_scale.net.api_client import (
     ApiClient,
-    backend_factory,
     with_api_errors_handling,
 )
-from neptune_scale.sync.errors_tracking import ErrorsQueue
 from neptune_scale.sync.parameters import (
     HTTP_REQUEST_MAX_TIME_SECONDS,
     MAX_REQUEST_SIZE_BYTES,
@@ -102,8 +99,6 @@ from neptune_scale.sync.parameters import (
 )
 from neptune_scale.util import (
     Daemon,
-    SharedFloat,
-    SharedInt,
     envs,
     get_logger,
 )
@@ -142,12 +137,6 @@ CODE_TO_ERROR: dict[IngestCode.ValueType, Optional[type[Exception]]] = {
 }
 
 
-class StatusTrackingElement(NamedTuple):
-    sequence_id: SequenceId
-    timestamp: datetime.datetime
-    request_id: str
-
-
 def code_to_exception(code: IngestCode.ValueType) -> Exception:
     if code in CODE_TO_ERROR:
         error = CODE_TO_ERROR[code]
@@ -155,72 +144,33 @@ def code_to_exception(code: IngestCode.ValueType) -> Exception:
     return NeptuneUnexpectedError(reason=f"Unexpected ingestion error code: {code}")
 
 
-class PeekableQueue(Generic[T]):
-    def __init__(self) -> None:
-        self._lock: threading.RLock = threading.RLock()
-        self._queue: queue.Queue[T] = queue.Queue()
-
-    def put(self, element: T) -> None:
-        with self._lock:
-            self._queue.put(element)
-
-    def peek(self, max_size: int) -> Optional[list[T]]:
-        with self._lock:
-            size = self._queue.qsize()
-            if size == 0:
-                return None
-
-            items = []
-            for i in range(min(size, max_size)):
-                item = self._queue.queue[i]
-                items.append(item)
-            return items
-
-    def commit(self, n: int) -> None:
-        with self._lock:
-            size = self._queue.qsize()
-            for _ in range(min(size, n)):
-                self._queue.get()
-
-
 def run_sync_process(
     operations_repository_path: Path,
-    errors_queue: ErrorsQueue,
     api_token: str,
     project: str,
     family: str,
-    last_queued_seq: SharedInt,
-    last_ack_seq: SharedInt,
-    last_ack_timestamp: SharedFloat,
 ) -> None:
     logger.info("Data synchronization started")
     stop_event = threading.Event()
     signal.signal(signal.SIGTERM, ft.partial(_handle_stop_signal, stop_event))
 
-    status_tracking_queue: PeekableQueue[StatusTrackingElement] = PeekableQueue()
     operations_repository = OperationsRepository(db_path=operations_repository_path)
+    operations_repository.delete_operation_submissions(up_to_seq_id=None)
+
     sender_thread = SenderThread(
         api_token=api_token,
         operations_repository=operations_repository,
-        status_tracking_queue=status_tracking_queue,
-        errors_queue=errors_queue,
         family=family,
-        last_queued_seq=last_queued_seq,
     )
     status_thread = StatusTrackingThread(
         api_token=api_token,
         project=project,
         operations_repository=operations_repository,
-        errors_queue=errors_queue,
-        status_tracking_queue=status_tracking_queue,
-        last_ack_seq=last_ack_seq,
-        last_ack_timestamp=last_ack_timestamp,
     )
     file_uploader_thread = FileUploaderThread(
         api_token=api_token,
         project=project,
         operations_repository=operations_repository,
-        errors_queue=errors_queue,
     )
     threads = [sender_thread, status_thread, file_uploader_thread]
     parent_process = psutil.Process(os.getpid()).parent()
@@ -234,9 +184,9 @@ def run_sync_process(
 
         # we're interrupting the status thread only if
         # - it's alive, and
-        # - the status tracking queue is empty, and
+        # - the operation statuses in operations repository are empty, and
         # - we know it will stay empty (since sender_thread is dead)
-        elif status_thread.is_alive() and status_tracking_queue.peek(max_size=1) is None:
+        elif status_thread.is_alive() and operations_repository.get_operation_submission_count(limit=1) == 0:
             status_thread.interrupt()
 
     def file_uploader_thread_died() -> bool:
@@ -310,18 +260,19 @@ class SenderThread(Daemon):
         api_token: str,
         family: str,
         operations_repository: OperationsRepository,
-        status_tracking_queue: PeekableQueue[StatusTrackingElement],
-        errors_queue: ErrorsQueue,
-        last_queued_seq: SharedInt,
     ) -> None:
         super().__init__(name="SenderThread", sleep_time=SYNC_THREAD_SLEEP_TIME)
 
         self._api_token: str = api_token
         self._family: str = family
         self._operations_repository: OperationsRepository = operations_repository
-        self._status_tracking_queue: PeekableQueue[StatusTrackingElement] = status_tracking_queue
-        self._errors_queue: ErrorsQueue = errors_queue
-        self._last_queued_seq: SharedInt = last_queued_seq
+
+        queued_range = operations_repository.get_operation_submission_sequence_id_range()
+        if queued_range is None:
+            last_queued_seq = SequenceId(-1)
+        else:
+            last_queued_seq = queued_range[1]
+        self._last_queued_seq: SequenceId = last_queued_seq
 
         self._backend: Optional[ApiClient] = None
         self._metadata: Metadata = operations_repository.get_metadata()  # type: ignore
@@ -330,7 +281,7 @@ class SenderThread(Daemon):
     @with_api_errors_handling
     def submit(self, *, operation: RunOperation) -> Optional[SubmitResponse]:
         if self._backend is None:
-            self._backend = backend_factory(api_token=self._api_token)
+            self._backend = ApiClient(api_token=self._api_token)
 
         response = self._backend.submit(operation=operation, family=self._family)
 
@@ -338,7 +289,7 @@ class SenderThread(Daemon):
         if status_code != 200:
             _raise_exception(status_code)
 
-        return response.parsed
+        return SubmitResponse.FromString(response.content)
 
     def work(self) -> None:
         try:
@@ -348,11 +299,9 @@ class SenderThread(Daemon):
                 - len(self._metadata.project)
                 - 200  # 200 bytes for RunOperation overhead,
             )
-            with self._last_queued_seq:
-                sequence_id = SequenceId(self._last_queued_seq.value)
 
             while operations := self._operations_repository.get_operations(
-                from_exclusive=sequence_id, up_to_bytes=max_operations_size
+                from_exclusive=self._last_queued_seq, up_to_bytes=max_operations_size
             ):
                 partitioned_operations = _partition_by_type_and_size(
                     operations, self._metadata.run_id, self._metadata.project, max_operations_size
@@ -361,7 +310,7 @@ class SenderThread(Daemon):
                 logger.debug(
                     "Start: submit %d RunOperations. Last queued seq: #%d",
                     len(partitioned_operations),
-                    sequence_id,
+                    self._last_queued_seq,
                 )
 
                 for run_operation, sequence_id, timestamp in partitioned_operations:
@@ -373,33 +322,30 @@ class SenderThread(Daemon):
 
                         last_request_id = request_ids.request_ids[-1]
 
-                        self._status_tracking_queue.put(
-                            StatusTrackingElement(
-                                sequence_id=sequence_id, request_id=last_request_id, timestamp=timestamp
-                            )
+                        self._operations_repository.save_operation_submissions(
+                            [
+                                OperationSubmission(
+                                    sequence_id=sequence_id,
+                                    request_id=last_request_id,
+                                    timestamp=int(timestamp.timestamp() * 1000),
+                                )
+                            ]
                         )
 
-                        # Update Last PUT sequence id and notify threads in the main process
-                        with self._last_queued_seq:
-                            self._last_queued_seq.value = sequence_id
-                            self._last_queued_seq.notify_all()
+                        self._last_queued_seq = sequence_id
                     except NeptuneRetryableError as e:
-                        self._errors_queue.put(e)
+                        self._operations_repository.save_errors([e], sequence_id=sequence_id)
                         # Sleep before retry
                         return
 
-                if logger.isEnabledFor(logging.DEBUG):
-                    # Don't access multiprocessing.Value if not in debug mode
-                    logger.debug(
-                        "Done: submit %d RunOperations. Last queued seq: #%d",
-                        len(partitioned_operations),
-                        self._last_queued_seq.value,
-                    )
+                logger.debug(
+                    "Done: submit %d RunOperations. Last queued seq: #%d",
+                    len(partitioned_operations),
+                    self._last_queued_seq,
+                )
 
         except Exception as e:
-            self._errors_queue.put(e)
-            with self._last_queued_seq:
-                self._last_queued_seq.notify_all()
+            self._operations_repository.save_errors([e])
             self.interrupt()
             raise NeptuneSynchronizationStopped() from e
 
@@ -472,20 +418,12 @@ class StatusTrackingThread(Daemon):
         api_token: str,
         project: str,
         operations_repository: OperationsRepository,
-        errors_queue: ErrorsQueue,
-        status_tracking_queue: PeekableQueue[StatusTrackingElement],
-        last_ack_seq: SharedInt,
-        last_ack_timestamp: SharedFloat,
     ) -> None:
         super().__init__(name="StatusTrackingThread", sleep_time=STATUS_TRACKING_THREAD_SLEEP_TIME)
 
         self._api_token: str = api_token
         self._project: str = project
         self._operations_repository: OperationsRepository = operations_repository
-        self._errors_queue: ErrorsQueue = errors_queue
-        self._status_tracking_queue: PeekableQueue[StatusTrackingElement] = status_tracking_queue
-        self._last_ack_seq: SharedInt = last_ack_seq
-        self._last_ack_timestamp: SharedFloat = last_ack_timestamp
 
         self._backend: Optional[ApiClient] = None
 
@@ -493,17 +431,11 @@ class StatusTrackingThread(Daemon):
         if self._backend is not None:
             self._backend.close()
 
-    def get_next(self) -> Optional[list[StatusTrackingElement]]:
-        try:
-            return self._status_tracking_queue.peek(max_size=MAX_REQUESTS_STATUS_BATCH_SIZE)
-        except queue.Empty:
-            return None
-
     @backoff.on_exception(backoff.expo, NeptuneRetryableError, max_time=HTTP_REQUEST_MAX_TIME_SECONDS)
     @with_api_errors_handling
     def check_batch(self, *, request_ids: list[str]) -> Optional[BulkRequestStatus]:
         if self._backend is None:
-            self._backend = backend_factory(api_token=self._api_token)
+            self._backend = ApiClient(api_token=self._api_token)
 
         response = self._backend.check_batch(request_ids=request_ids, project=self._project)
 
@@ -512,26 +444,25 @@ class StatusTrackingThread(Daemon):
         if status_code != 200:
             _raise_exception(status_code)
 
-        return response.parsed
+        return BulkRequestStatus.FromString(response.content)
 
     def work(self) -> None:
         try:
-            while (batch := self.get_next()) is not None:
+            while batch := self._operations_repository.get_operation_submissions(limit=MAX_REQUESTS_STATUS_BATCH_SIZE):
                 request_ids = [element.request_id for element in batch]
                 sequence_ids = [element.sequence_id for element in batch]
-                timestamps = [element.timestamp for element in batch]
 
                 try:
                     response = self.check_batch(request_ids=request_ids)
                     if response is None:
                         raise NeptuneUnexpectedError("Server response is empty")
                 except NeptuneRetryableError as e:
-                    self._errors_queue.put(e)
+                    self._operations_repository.save_errors([e])
                     # Small give up, sleep before retry
                     break
 
-                operations_to_commit, processed_sequence_id, processed_timestamp, fatal_sync_error = 0, None, None, None
-                for request_status, request_sequence_id, timestamp in zip(response.statuses, sequence_ids, timestamps):
+                processed_sequence_id, fatal_sync_error = None, None
+                for request_status, request_sequence_id in zip(response.statuses, sequence_ids):
                     if any(code_status.code == Code.UNAVAILABLE for code_status in request_status.code_by_count):
                         logger.debug(f"Operation #{request_sequence_id} is not yet processed.")
                         # Request status not ready yet, sleep and retry
@@ -547,42 +478,31 @@ class StatusTrackingThread(Daemon):
                     if fatal_sync_error is not None:
                         break
 
-                    for error in errors:
-                        self._errors_queue.put(error)
+                    if errors:
+                        self._operations_repository.save_errors(
+                            errors=errors,
+                            sequence_id=request_sequence_id,
+                        )
 
-                    operations_to_commit += 1
-                    processed_sequence_id, processed_timestamp = request_sequence_id, timestamp
+                    processed_sequence_id = request_sequence_id
 
-                if operations_to_commit > 0:
-                    self._status_tracking_queue.commit(operations_to_commit)
+                if processed_sequence_id is not None:
+                    logger.debug(f"Operations up to #{processed_sequence_id} are completed.")
 
-                    # Update Last ACK sequence id and notify threads in the main process
-                    if processed_sequence_id is not None:
-                        logger.debug(f"Operations up to #{processed_sequence_id} are completed.")
-
-                        self._operations_repository.delete_operations(up_to_seq_id=processed_sequence_id)
-
-                        with self._last_ack_seq:
-                            self._last_ack_seq.value = processed_sequence_id
-                            self._last_ack_seq.notify_all()
-
-                    # Update Last ACK timestamp and notify threads in the main process
-                    if processed_timestamp is not None:
-                        with self._last_ack_timestamp:
-                            self._last_ack_timestamp.value = processed_timestamp.timestamp()
-                            self._last_ack_timestamp.notify_all()
+                    # TODO: delete in a single transaction
+                    self._operations_repository.delete_operations(up_to_seq_id=processed_sequence_id)
+                    self._operations_repository.delete_operation_submissions(up_to_seq_id=processed_sequence_id)
 
                 if fatal_sync_error is not None:
                     raise fatal_sync_error
 
-                if operations_to_commit == 0:
+                if processed_sequence_id is None:
                     # Sleep before retry
                     break
 
         except Exception as e:
-            self._errors_queue.put(e)
+            self._operations_repository.save_errors([e])
             self.interrupt()
-            self._last_ack_seq.notify_all()
             raise NeptuneSynchronizationStopped() from e
 
     @staticmethod
@@ -596,7 +516,6 @@ class FileUploaderThread(Daemon):
         project: str,
         api_token: str,
         operations_repository: OperationsRepository,
-        errors_queue: ErrorsQueue,
         max_concurrent_uploads: Optional[int] = None,
     ) -> None:
         super().__init__(name="FileUploaderThread", sleep_time=1)
@@ -604,10 +523,14 @@ class FileUploaderThread(Daemon):
         self._project = project
         self._neptune_api_token = api_token
         self._operations_repository = operations_repository
-        self._errors_queue = errors_queue
         self._max_concurrent_uploads = max_concurrent_uploads or envs.get_positive_int(
             envs.MAX_CONCURRENT_FILE_UPLOADS, 50
         )
+        self._upload_chunk_size = envs.get_positive_int("NEPTUNE_FILE_UPLOAD_CHUNK_SIZE", 16 * 1024 * 1024)
+        logger.debug(f"{self._upload_chunk_size=}")
+        # This is enforced by GCP, enforcing this for other providers is consistent.
+        if self._upload_chunk_size % (256 * 1024) != 0:
+            raise ValueError("NEPTUNE_FILE_UPLOAD_CHUNK_SIZE must be a multiple of 256 KiB")
 
         self._api_client: Optional[ApiClient] = None
 
@@ -618,7 +541,7 @@ class FileUploaderThread(Daemon):
     def work(self) -> None:
         try:
             if self._api_client is None:
-                self._api_client = backend_factory(api_token=self._neptune_api_token)
+                self._api_client = ApiClient(api_token=self._neptune_api_token)
                 asyncio.set_event_loop(self._aio_loop)
 
             while file_upload_requests := self._operations_repository.get_file_upload_requests(
@@ -633,32 +556,42 @@ class FileUploaderThread(Daemon):
                 # Note that self._upload_file() should not raise an exception, as they are handled
                 # in the method itself. However, we still pass return_exceptions=True to make sure asyncio.gather()
                 # waits for all the tasks to finish regardless of any exceptions.
-                tasks = [self._upload_file(file, storage_urls[file.destination]) for file in file_upload_requests]
+                tasks = []
+                for file in file_upload_requests:
+                    provider, storage_url = storage_urls[file.destination]
+                    tasks.append(self._upload_file(file, provider, storage_url))
+
                 self._aio_loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
 
                 logger.debug("Upload tasks completed for the current batch")
         except NeptuneRetryableError as e:
-            self._errors_queue.put(e)
+            self._operations_repository.save_errors([e])
         except Exception as e:
             logger.error("Fatal error in file uploader thread", exc_info=e)
-            self._errors_queue.put(e)
+            self._operations_repository.save_errors([e])
             self.interrupt()
             raise NeptuneSynchronizationStopped() from e
 
-    async def _upload_file(self, file: FileUploadRequest, storage_url: str) -> None:
+    async def _upload_file(self, file: FileUploadRequest, provider: str, storage_url: str) -> None:
         try:
-            await upload_file(file.source_path, file.mime_type, storage_url)
+            if provider == Provider.AZURE:
+                await upload_to_azure(file.source_path, file.mime_type, storage_url, chunk_size=self._upload_chunk_size)
+            elif provider == Provider.GCP:
+                await upload_to_gcp(file.source_path, file.mime_type, storage_url, chunk_size=self._upload_chunk_size)
+            else:
+                raise NeptuneUnexpectedError(f"Unsupported file storage provider: {provider}")
+
             if file.is_temporary:
                 logger.debug(f"Removing temporary file {file.source_path}")
                 Path(file.source_path).unlink(missing_ok=True)
 
             self._operations_repository.delete_file_upload_requests([file.sequence_id])  # type: ignore
         except NeptuneRetryableError as e:
-            self._errors_queue.put(e)
+            self._operations_repository.save_errors([e])
         except Exception as e:
             # Fatal failure. Do not retry the file, but keep it on disk.
             logger.error(f"Error while uploading file {file.source_path}", exc_info=e)
-            self._errors_queue.put(NeptuneFileUploadError())
+            self._operations_repository.save_errors([NeptuneFileUploadError()])
             self._operations_repository.delete_file_upload_requests([file.sequence_id])  # type: ignore
 
     def close(self) -> None:
@@ -671,28 +604,37 @@ class FileUploaderThread(Daemon):
 
 @backoff.on_exception(backoff.expo, NeptuneRetryableError, max_time=HTTP_REQUEST_MAX_TIME_SECONDS)
 @with_api_errors_handling
-def fetch_file_storage_urls(client: ApiClient, project: str, destination_paths: list[str]) -> dict[str, str]:
-    """Fetch Azure urls for storing files. Return a dict of target_path -> upload url"""
+def fetch_file_storage_urls(
+    client: ApiClient, project: str, destination_paths: list[str]
+) -> dict[str, tuple[str, str]]:
+    """Fetch signed URLs for storing files. Returns a dict of target_path -> (provider, upload url)."""
+
     logger.debug("Fetching file storage urls")
     response = client.fetch_file_storage_urls(paths=destination_paths, project=project, mode="write")
     status_code = response.status_code
     if status_code != 200:
+        logger.debug(f"{response.content=!r}")
         _raise_exception(status_code)
 
     if response.parsed is None:
         raise NeptuneUnexpectedResponseError("Server response is empty")
 
-    return {file.path: file.url for file in response.parsed.files}
+    return {file.path: (file.provider, file.url) for file in response.parsed.files or []}
 
 
-async def upload_file(local_path: str, mime_type: str, storage_url: str) -> None:
-    logger.debug(f"Start: upload file {local_path}")
+async def upload_to_azure(local_path: str, mime_type: str, storage_url: str, chunk_size: int = 4 * 1024 * 1024) -> None:
+    logger.debug(f"Starting upload to Azure: {local_path}, {mime_type=}, {chunk_size=}")
 
     try:
         size_bytes = Path(local_path).stat().st_size
         with open(local_path, "rb") as file:
             client = BlobClient.from_blob_url(
-                storage_url, initial_backoff=5, increment_base=3, retry_total=5, transport=AsyncioRequestsTransport()
+                storage_url,
+                max_block_size=chunk_size,
+                max_initial_backoff=5,
+                increment_base=3,
+                retry_total=5,
+                transport=AsyncioRequestsTransport(),
             )
             async with client:
                 # Upload with default concurrency settings
@@ -706,6 +648,7 @@ async def upload_file(local_path: str, mime_type: str, storage_url: str) -> None
         logger.debug(f"Azure SDK error, will retry uploading file {local_path}: {e}")
         raise NeptuneFileUploadTemporaryError() from e
     except Exception as e:
+        logger.debug(f"Failed to upload file {local_path}: {e}")
         raise e
 
-    logger.debug(f"Done: upload file {local_path}")
+    logger.debug(f"Finished upload to Azure: {local_path}")
